import globalvars
from read_data import read_data
from initialization import initialization
from assembly import assembly
from solver import solve
from postprocess import postprocess


def main():
    read_data()
    initialization()
    assembly()
    solve()
    postprocess()


if __name__ == '__main__':
    main()
    


import globalvars
import numpy as np
import math
from initialization import initialization
from  gauss_quadrature import GaussQuadrature, Set_Ngauss
from dmat import dmat_Solids2D

def GiveNdime(ElemType):
    if(ElemType == "BAR02" or ElemType == "BAR03"):
        ndime=1
    elif(ElemType == "TRUSS02" or ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
      ndime=2
    return ndime

def GiveNComp(ElemType, ProblemType):
  if(ProblemType == "Structural_Mechanics"):
    if(ElemType == "BAR02" or ElemType == "BAR03" or ElemType == "TRUSS02"):
        ncomp=1
    elif(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
        ncomp = 3
  elif (ProblemType == "Thermal"):
    if(ElemType == "BAR02" or ElemType == "BAR03" or ElemType == "TRUSS02"):
      ncomp=1
    elif(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
      ncomp = 2
  return ncomp

def GiveNdof(ElemType, ProblemType):
  if(ProblemType == "Structural_Mechanics"):
    if(ElemType == "BAR02" or ElemType == "BAR03"):
      ndof=1
    elif(ElemType == "TRUSS02" or ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
      ndof=2
  elif (ProblemType == "Thermal"):
    if(ElemType == "BAR02" or ElemType == "BAR03"):
      ndof=1
    elif(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
      ndof=2
  return ndof

def GiveNnodes(ElemType):
  if(ElemType == "BAR02"):
    nnodes = 2
  elif(ElemType == "BAR03"):
    nnodes = 3
  elif(ElemType == "TR03"):
    nnodes = 3
  elif(ElemType == "TR06"):
    nnodes = 6
  elif(ElemType == "QU04"):
    nnodes = 4
  elif(ElemType == "QU08"):
    nnodes = 8
  elif(ElemType == "QU09"):
    nnodes = 9
  return nnodes

def FuncForm(ElemType, pos_pg):
  ndime = GiveNdime(ElemType)
  if(ndime == 1):
    s = pos_pg[0]
    t = 0
  elif(ndime == 2):
    s = pos_pg[0]
    t = pos_pg[1]

  nnodes = GiveNnodes(ElemType)
  fform = np.zeros((nnodes), dtype=float)
  if(ElemType == "BAR02" or ElemType == "TRUSS02"):
    fform[0] = 0.5 * (1 - s)
    fform[1] = 0.5 * (1 + s)
  elif(ElemType == "BAR03"):
    fform[0] = 0.5 * s * (s - 1)
    fform[1] = (1 + s) * (1 - s)
    fform[2] = 0.5 * s * (1 + s)
  elif(ElemType == "TR03"):
    fform[0] = 1 - s - t
    fform[1] = s
    fform[2] = t
  elif(ElemType == "TR06"):
    st = s * t
    tt = t * t
    ss = s * s
    fform[0] = 1 - (s + t) * 3 + (ss + tt) * 2 + st * 4
    fform[1] = (s - ss - st) * 4
    fform[2] = ss * 2 - s
    fform[3] = 4 * st
    fform[4] = tt * 2 - t
    fform[5] = (t - tt - st) * 4
  elif(ElemType == "QU04"):
    st = s * t
    fform[0] = (1 - s - t + st) / 4
    fform[1] = (1 + s - t - st) / 4
    fform[2] = (1 + s + t + st) / 4
    fform[3] = (1 - s + t - st) / 4
  elif(ElemType == "QU08"):
    st = s * t
    tt = t * t
    ss = s * s
    sst = ss * t
    stt = tt * s
    fform[0] = (-1 + st + ss + tt - sst - stt) * 0.25
    fform[1] = (1 - t - ss + sst) * 0.5
    fform[2] = (-1 - st + ss + tt - sst + stt) * 0.25
    fform[3] = (1 + s - tt - stt) * 0.5
    fform[4] = (-1 + st + ss + tt + sst + stt) * 0.25
    fform[5] = (1 + t - ss - sst) * 0.5
    fform[6] = (-1 - st + ss + tt + sst - stt) * 0.25
    fform[7] = (1 - s - tt + stt) * 0.5
  elif(ElemType == "QU09"):
    st = s * t
    tt = t * t
    ss = s * s
    s9 = s - 1
    t9 = t - 1
    s1 = s + 1
    t1 = t + 1
    fform[0] = 0.25 * s9 * st * t9
    fform[1] = 0.5 * (1 - ss) * t * t9
    fform[2] = 0.25 * s1 * st * t9
    fform[3] = 0.5 * s * s1 * (1 - tt)
    fform[4] = 0.25 * s1 * st * t1
    fform[5] = 0.5 * (1 - ss) * t * t1
    fform[6] = 0.25 * s9 * st * t1
    fform[7] = 0.5 * s * s9 * (1 - tt)
    fform[8] = (1 - ss) * (1 - tt)
  return fform

def DerivNatCoord(ElemType, pos_pg):
  ndime = GiveNdime(ElemType)
  if(ndime == 1):
    s = pos_pg[0]
    t = 0
  elif(ndime == 2):
    s = pos_pg[0]
    t = pos_pg[1]

  nnodes = GiveNnodes(ElemType)
  deriv = np.zeros((ndime,nnodes), dtype=float)

  #deriv[ider,inode]
  if(ElemType == "BAR02" or ElemType == "TRUSS02"):
    deriv[0,0] =-0.5
    deriv[0, 1] = 0.5
  elif(ElemType == "BAR03"):
    deriv[0,0] = s - 0.5
    deriv[0,1] = -2 * s
    deriv[0, 2] = s - 0.5
  elif(ElemType == "TR03"):
    deriv[0,0] = -1
    deriv[0,1] =  1
    deriv[0,2] =  0
    deriv[1,0] = -1
    deriv[1,1] =  0
    deriv[1, 2] = -1
  elif(ElemType == "TR06"):
    deriv[0,0]=-3+(s+t)*4
    deriv[0,1]=(1-s*2-t)*4
    deriv[0,2]=s*4-1
    deriv[0,3]=t*4
    deriv[0,4]=0
    deriv[0,5]=-t*4
    deriv[1,0]=-3+(t+s)*4
    deriv[1,1]=-s*4
    deriv[1,2]=0
    deriv[1,3]=4*s
    deriv[1,4]=t*4-1
    deriv[1,5]=(1-t*2-s)*4
  elif(ElemType == "QU04"):
    deriv[0,0] = (-1 + t) / 4
    deriv[0,1] = (1 - t) / 4
    deriv[0,2] = (1 + t) / 4
    deriv[0,3] = (-1 - t) / 4
    deriv[1,0] = (-1 + s) / 4
    deriv[1,1] = (-1 - s) / 4
    deriv[1,2] = (1 + s) / 4
    deriv[1,3] = (1 - s) / 4
  elif(ElemType == "QU08"):
    s2 = 2 * s
    t2 = 2 * t
    st = s * t
    st2 = s * t * 2
    tt = t * t
    ss = s * s
    deriv[0,0] = (t + s2 - st2 - tt) * 0.25
    deriv[0,1] = -s + st
    deriv[0,2] = (-t + s2 - st2 + tt) * 0.25
    deriv[0,3] = (1.0 - tt) * 0.5
    deriv[0,4] = (t + s2 + st2 + tt) * 0.25
    deriv[0,5] = -s - st
    deriv[0,6] = (-t + s2 + st2 - tt) * 0.25
    deriv[0,7] = (-1.0 + tt) * 0.5
    deriv[1,0] = (s + t2 - ss - st2) * 0.25
    deriv[1,1] = (-1.0 + ss) * 0.5
    deriv[1,2] = (-s + t2 - ss + st2) * 0.25
    deriv[1,3] = -t - st
    deriv[1,4] = (s + t2 + ss + st2) * 0.25
    deriv[1,5] = (1.0 - ss) * 0.5
    deriv[1,6] = (-s + t2 + ss - st2) * 0.25
    deriv[1,7] = -t + st
  elif(ElemType == "QU09"):
    s2 = 2 * s
    t2 = 2 * t
    st = s * t
    st2 = s * t * 2
    tt = t * t
    ss = s * s
    s9 = s - 1
    t9 = t - 1
    s1 = s + 1
    t1 = t + 1
    deriv[0,0] = 0.25 * t * t9 * (-1 + s2)
    deriv[0,1] = -st * t9
    deriv[0,2] = 0.25 * (1 + s2) * t * t9
    deriv[0,3] = 0.5 * (1 + s2) * (1 - tt)
    deriv[0,4] = 0.25 * (1 + s2) * t * t1
    deriv[0,5] = -st * t1
    deriv[0,6] = 0.25 * (-1 + s2) * t * t1
    deriv[0,7] = 0.5 * (-1 + s2) * (1 - tt)
    deriv[0,8] = -s2 * (1 - tt)
    deriv[1,0] = 0.25 * (-1 + t2) * s * s9
    deriv[1,1] = 0.5 * (1 - ss) * (-1 + t2)
    deriv[1,2] = 0.25 * s * s1 * (-1 + t2)
    deriv[1,3] = -st * s1
    deriv[1,4] = 0.25 * s * s1 * (1 + t2)
    deriv[1,5] = 0.5 * (1 - ss) * (1 + t2)
    deriv[1,6] = 0.25 * s * s9 * (1 + t2)
    deriv[1,7] = -st * s9
    deriv[1,8] = -t2 * (1 - ss)
  return deriv

def det_Jacob(elem, ElemType, Nodes, pos_pg): 
  ndime = GiveNdime(ElemType)
  nnodes = elem.give_Nnodes()
  corel = np.zeros((nnodes,ndime), dtype=float)
  for inode in range(nnodes):
    id_node = elem["Connectivities"][inode] - 1
    for idime in range(ndime):
      if(idime == 0):
        corel[inode, idime] = Nodes[id_node]["x"]
      elif (idime == 1):
        corel[inode, idime] = Nodes[id_node]["y"]

  deriv_nat = DerivNatCoord(ElemType, pos_pg)
  jacobMat = np.matmul(deriv_nat,corel)
  det_jacobMat = np.linalg.det(jacobMat)
  if (det_jacobMat==0):
    raise NameError("Error: Jacobian negative")
  return det_jacobMat

def derivCartesian(elem, ElemType, Nodes, pos_pg):
  ndime = GiveNdime(ElemType)
  nnodes = GiveNnodes(ElemType)
  corel = np.zeros((nnodes,ndime), dtype=float)
  for inode in range(nnodes):
    id_node = elem["Connectivities"][inode] - 1
    for idime in range(ndime):
      if(idime == 0):
        corel[inode, idime] = Nodes[id_node]["x"]
      elif (idime == 1):
        corel[inode, idime] = Nodes[id_node]["y"]

  deriv_nat = DerivNatCoord(ElemType, pos_pg)
  jacobMat = np.matmul(deriv_nat, corel)
  det_jacobMat = np.linalg.det(jacobMat)
  if (det_jacobMat==0):
    raise NameError("Error: Jacobian negative")
  JacobMat_inv = np.linalg.inv(jacobMat)
  derivCart = np.matmul(JacobMat_inv, deriv_nat)
  return det_jacobMat, derivCart
 

def Smooth_Strains():
    num_nodes = len(globalvars.data["Mesh"]["Nodes"])
    nodal_Eps = np.zeros((num_nodes, 3), dtype=float)
    node_counter = np.zeros((num_nodes), dtype=int)
    ProblemType = globalvars.data["ProblemType"]
    ElemType = globalvars.data["Mesh"]["ElemType"]
    nnode = GiveNnodes(ElemType)
    ngauss = Set_Ngauss(ElemType)
    ncomp = GiveNComp(ElemType, ProblemType)
    Eps_vec = np.zeros((ncomp, ngauss), dtype=float)
    num_elems = len(globalvars.data["Mesh"]["Elements"])
    for icomp in range(ncomp):
      globalvars.Eps_elem_vec[icomp] = np.zeros((num_elems, ngauss), dtype=float)
    ielem = 0
    for elem in globalvars.data["Mesh"]["Elements"]:
        if(ElemType == "TR03"):
          Esp_vec_TR03 = Strain_TR03(elem)
          Eps_vec[:ncomp,:] = Esp_vec_TR03[:ncomp]
        else:
          for igauss in range(ngauss):
            Eps_vec[:, igauss] = Strain_Solid(elem, ElemType, ProblemType, igauss)

        for igauss in range(ngauss):
          for icomp in range(ncomp):
            globalvars.Eps_elem_vec[icomp][ielem,igauss] = Eps_vec[icomp, igauss]
        ielem +=1
        
        for inode in range(nnode):
            id_node = elem["Connectivities"][inode] - 1
            node_counter[id_node] += 1
            nodal_Eps[id_node,:ncomp] += Eps_vec[:ncomp, inode]
    
    for inode in range(num_nodes):
        nodal_Eps[inode,:ncomp] /= node_counter[inode]

    return nodal_Eps

def Smooth_Stresses():
    num_nodes = len(globalvars.data["Mesh"]["Nodes"])
    nodal_Sigma = np.zeros((num_nodes, 3), dtype=float)
    node_counter = np.zeros((num_nodes), dtype=int)
    ProblemType = globalvars.data["ProblemType"]
    ElemType = globalvars.data["Mesh"]["ElemType"]
    nnode = GiveNnodes(ElemType)
    ngauss = Set_Ngauss(ElemType)
    ncomp = GiveNComp(ElemType, ProblemType)
    Eps_vec = np.zeros((ncomp, ngauss), dtype=float)
    Sigma_vec = np.zeros((ncomp, ngauss), dtype=float)
    num_elems = len(globalvars.data["Mesh"]["Elements"])
    for icomp in range(ncomp):
      globalvars.Sigma_elem_vec[icomp] = np.zeros((num_elems, ngauss), dtype=float)
    ielem = 0
    for elem in globalvars.data["Mesh"]["Elements"]:
        if (ElemType == "TR03"):
          Sigma_vec_TR03 = Stress_TR03(elem)
          Sigma_vec[:ncomp,:] = Sigma_vec_TR03[:ncomp]
        else:
          for igauss in range(ngauss):
            if(len(globalvars.Eps_elem_vec[0]) == 0):
              Eps_vec[:, igauss] = Strain_Solid(elem, ElemType, ProblemType, igauss)
            else:
              for icomp in range(ncomp):
                Eps_vec[icomp, igauss] = globalvars.Eps_elem_vec[icomp][ielem,igauss]
            Sigma_vec[:, igauss] = Stress_Solid(elem, ElemType, Eps_vec[:, igauss])

        for igauss in range(ngauss):
          for icomp in range(ncomp):
            globalvars.Sigma_elem_vec[icomp][ielem, igauss] = Sigma_vec[icomp, igauss]
        ielem += 1
        
        for inode in range(nnode):
            id_node = elem["Connectivities"][inode] - 1
            node_counter[id_node] += 1
            nodal_Sigma[id_node,:ncomp] += Sigma_vec[:ncomp, inode]
    
    for inode in range(num_nodes):
        nodal_Sigma[inode,:ncomp] /= node_counter[inode]

    return nodal_Sigma


def Smooth_Forces():
    num_nodes = len(globalvars.data["Mesh"]["Nodes"])
    nodal_Forces = np.zeros((num_nodes, 3), dtype=float)
    node_counter = np.zeros((num_nodes), dtype=int)
    ProblemType = globalvars.data["ProblemType"]
    ElemType = globalvars.data["Mesh"]["ElemType"]
    nnode = GiveNnodes(ElemType)
    ngauss = Set_Ngauss(ElemType)
    ncomp = GiveNComp(ElemType, ProblemType)
    Eps_vec = np.zeros((ncomp, ngauss), dtype=float)
    Sigma_vec = np.zeros((ncomp, ngauss), dtype=float)
    Forces_vec = np.zeros((ncomp, ngauss), dtype=float)
    ielem = 0
    for elem in globalvars.data["Mesh"]["Elements"]:
        mat_id = elem["MaterialId"] - 1
        material = globalvars.data["Materials"][mat_id]
        factor = 1.0
        if(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
          if (material["Plane_Type"] == "Plane_Stress"):
            factor = material["Thickness"]
        elif (ElemType == "TRUSS02" or ElemType == "BAR02" or ElemType == "BAR03"):
          factor = material["Area"]
        if (ElemType == "TR03"):
          Sigma_vec_TR03 = Stress_TR03(elem)
          Forces_vec[:ncomp,:] = Sigma_vec_TR03[:ncomp]*factor
        else:
          for igauss in range(ngauss):
            if (len(globalvars.Sigma_elem_vec[0]) == 0):
              Eps_vec[:, igauss] = Strain_Solid(elem, ElemType, ProblemType, igauss)
              Sigma_vec[:, igauss] = Stress_Solid(elem, ElemType, Eps_vec[:, igauss])
            else:
              for icomp in range(ncomp):
                Sigma_vec[icomp, igauss] = globalvars.Sigma_elem_vec[icomp][ielem, igauss]
            Forces_vec[:, igauss] = Sigma_vec[:, igauss] * factor
        ielem += 1
        
        for inode in range(nnode):
            id_node = elem["Connectivities"][inode] - 1
            node_counter[id_node] += 1
            nodal_Forces[id_node,:ncomp] += Forces_vec[:ncomp, inode]
    
    for inode in range(num_nodes):
        nodal_Forces[inode,:ncomp] /= node_counter[inode]

    return nodal_Forces   

def Bmat_TR03(elem):
    x = np.zeros((3), dtype=float)
    y = np.zeros((3), dtype=float)
    b = np.zeros((3), dtype=float)
    c = np.zeros((3), dtype=float)
    Bmat = np.zeros((3,6), dtype=float)
    for inode in range(3):
        id_node = elem["Connectivities"][inode] - 1
        x[inode] = globalvars.data["Mesh"]["Nodes"][id_node]["x"]
        y[inode] = globalvars.data["Mesh"]["Nodes"][id_node]["y"]

    b[0] = y[1] - y[2]
    b[1] = y[2] - y[0]
    b[2] = y[0] - y[1]
    c[0] = x[2] - x[1]
    c[1] = x[0] - x[2]
    c[2] = x[1] - x[0]
    
    coords_Mat = np.array([[1,x[0],y[0]], [1,x[1],y[1]],[1,x[2],y[2]]]) 
    Area = 0.5 * np.linalg.det(coords_Mat)
    for inode in range(3):
        icol1 = inode * 2
        icol2 = inode * 2 + 1
        Bmat[0, icol1] = b[inode] / (2.0 * Area)
        Bmat[1, icol2] = c[inode] / (2.0 * Area)
        Bmat[2, icol1] = c[inode] / (2.0 * Area)
        Bmat[2, icol2] = b[inode] / (2.0 * Area)    

    return Bmat, Area

def Strain_TR03(elem):
    u_vec = np.zeros((6), dtype=float)
    for inode in range(3):
        id_node = elem["Connectivities"][inode] - 1
        for igl in range(2):
            idire = globalvars.madgln[id_node, igl]
            ipos = inode*2 + igl
            u_vec[ipos] = globalvars.u_vec[idire]
    
    Bmat, _ = Bmat_TR03(elem)
    Eps_vec = np.matmul(Bmat, u_vec)
    return Eps_vec

def Strain_Solid(elem, ElemType, ProblemType, igauss):
    nnodes = GiveNnodes(ElemType)
    ndof = globalvars.ndof
    uvec_size = nnodes * ndof
    u_vec = np.zeros((uvec_size), dtype=float)
    Nodes = globalvars.data["Mesh"]["Nodes"]
    for inode in range(nnodes):
        id_node = elem["Connectivities"][inode] - 1
        for igl in range(ndof):
            idire = globalvars.madgln[id_node, igl]
            ipos = inode*ndof + igl
            u_vec[ipos] = globalvars.u_vec[idire]
    
    if(ElemType == "TR03"):
      Bmat, _ = Bmat_TR03(elem)
    else:
      pos_pg, _ = GaussQuadrature(ElemType)
      _, derivCart = derivCartesian(elem, ElemType, Nodes, pos_pg[igauss])
      Bmat = Calc_Bmat(ElemType, derivCart, ProblemType)
  
    Eps_vec = np.matmul(Bmat, u_vec)
    return Eps_vec

def Stress_TR03(elem):
    mat_id = elem["MaterialId"] - 1
    material = globalvars.data["Materials"][mat_id]
    Dmat = dmat_Solids2D(material)
    Eps_vec = Strain_TR03(elem)
    Sigma_vec = np.matmul(Dmat, Eps_vec)
    return Sigma_vec

def Stress_Solid(elem, ElemType, Eps_vec):
    mat_id = elem["MaterialId"] - 1
    material = globalvars.data["Materials"][mat_id]
    if(ElemType == "BAR02" or ElemType == "BAR03" or ElemType == "TRUSS02"):
      Young = material["Young"]
      Sigma_vec = Young * Eps_vec
    else:
      Dmat = dmat_Solids2D(material)
      Sigma_vec = np.matmul(Dmat, Eps_vec)
    return Sigma_vec

def Calc_Bmat(ElemType, derivCart, ProblemType):
  ndime = GiveNdime(ElemType)
  nnodes = GiveNnodes(ElemType)
  if(ProblemType == "Thermal"):
    nrows = ndime
    ncols = nnodes
  elif(ProblemType == "Structural_Mechanics"):
    if (ndime == 1):
      nrows = 1
      ncols = nnodes
    elif(ndime == 2):
      nrows = 3 #This case not include revolution solid case
      ncols = nnodes * 2
      
  B_matrix = np.zeros((nrows, ncols), dtype=float)

  if(ProblemType == "Thermal"):
    for idime in range(ndime):
      for inode in range(nnodes):
        B_matrix[idime,inode] = derivCart[idime,inode]
  elif(ProblemType == "Structural_Mechanics"):
    if (ndime == 1):
      for idime in range(ndime):
        for inode in range(nnodes):
          B_matrix[idime,inode] = derivCart[idime,inode]
    elif(ndime == 2):
      for inode in range(nnodes):
        col_ini = inode * ndime
        B_matrix[0, col_ini] = derivCart[0, inode]
        B_matrix[1, col_ini + 1] = derivCart[1, inode]
        B_matrix[2, col_ini] = derivCart[1, inode]
        B_matrix[2, col_ini + 1] = derivCart[0, inode]
  return B_matrix

def NtxN(ElemType, pos_pg, ProblemType):
  fform = FuncForm(ElemType, pos_pg)
  nnodes = GiveNnodes(ElemType)
  ndime = GiveNnodes(ElemType)
 
  if(ProblemType == "Thermal"):
    nrows = nnodes
    ncols = nnodes
    ngl = 1
  elif(ProblemType == "Structural_Mechanics"):
    if (ndime == 1):
      nrows = nnodes
      ncols = nnodes
      ngl = 1
    elif(ndime == 2):
      nrows = nnodes * 2
      ncols = nnodes * 2
      ngl = 3

  NtxN = np.zeros((nrows,ncols), dtype=float)

  for inode in range(nnodes):
    for jnode in range(nnodes):
      for igl in range(ngl):
        irow = inode * ngl + igl
        icol = jnode * ngl + igl
        NtxN[irow,icol] = fform[inode] * fform[jnode]
  
  return NtxN





import globalvars
import numpy as np
import math
from initialization import initialization
from solids_utils import Bmat_TR03, Calc_Bmat
from dmat import dmat_Solids2D, giveLocalStiffness
from gauss_quadrature import Set_Ngauss, GaussQuadrature
from solids_utils import derivCartesian, GiveNdof, GiveNnodes

def assembly():
    stiff_assembly()
    loads_assembly()

def stiff_assembly():
    mesh_type = globalvars.data["Mesh"]["ElemType"]
    for elem in globalvars.data["Mesh"]["Elements"]:
        if (mesh_type == "BAR02"):
            rigimat = stiffness_BAR02(elem)
        elif (mesh_type == "TRUSS02"):
            rigimat = stiffness_TRUSS02(elem)
        elif (mesh_type == "TR03"):
            rigimat = stiffness_TR03(elem)
        elif (mesh_type == "QU04"):
            rigimat = stiffness_QU04(elem)
        
        conec_list = elem["Connectivities"]
        doflist = np.zeros((len(conec_list)*globalvars.ndof), dtype=int)
        idof = 0
        for inode in range(len(conec_list)):
            id_node = conec_list[inode]-1
            for igl in range(globalvars.ndof):
                doflist[idof] =  globalvars.madgln[id_node, igl]
                idof += 1
        
        assamk(doflist, rigimat)


def loads_assembly():
    Loads = globalvars.data["Loads"]

    if ("Point_Loads" in Loads):
        doflist = np.zeros((globalvars.ndof), dtype=int)
        fvect = np.zeros((globalvars.ndof), dtype=float)
        for pointLoad in Loads["Point_Loads"]:
            id_node = pointLoad["Node"] - 1
            idof = 0
            for igl in range(globalvars.ndof):
                doflist[idof] = globalvars.madgln[id_node, igl]
                fvect[idof] = pointLoad["Values"][igl]
                idof += 1
            assamf(doflist, fvect)   

    if ("Surface_Loads" in Loads):
        print("There is not surface loads yet...")

        
def stiffness_BAR02(elem):
    mat_id = elem["MaterialId"] - 1
    material = globalvars.data["Materials"][mat_id]
    node1 = elem["Connectivities"][0] - 1
    x1 = globalvars.data["Mesh"]["Nodes"][node1]["x"]
    y1 = globalvars.data["Mesh"]["Nodes"][node1]["y"]
    node2 = elem["Connectivities"][1] - 1
    x2 = globalvars.data["Mesh"]["Nodes"][node2]["x"]
    y2 = globalvars.data["Mesh"]["Nodes"][node2]["y"]
    l = math.sqrt((x2 - x1)** 2 + (y2 - y1)** 2)
    k = giveLocalStiffness(material, l)
    rigimat = np.zeros((2, 2), dtype=float)
    rigimat[0, 0] = k
    rigimat[0, 1] = -k
    rigimat[1, 0] = -k
    rigimat[1, 1] = k
    return rigimat


def stiffness_TRUSS02(elem):
    mat_id = elem["MaterialId"] - 1
    material = globalvars.data["Materials"][mat_id]
    Young = material["Young"]
    Area = material["Area"]
    node1 = elem["Connectivities"][0] - 1
    x1 = globalvars.data["Mesh"]["Nodes"][node1]["x"]
    y1 = globalvars.data["Mesh"]["Nodes"][node1]["y"]
    node2 = elem["Connectivities"][1] - 1
    x2 = globalvars.data["Mesh"]["Nodes"][node2]["x"]
    y2 = globalvars.data["Mesh"]["Nodes"][node2]["y"]
    length = math.sqrt((x2 - x1)** 2 + (y2 - y1)** 2)
    cos_alpha = (x2 - x1) / length
    sin_alpha = (y2 - y1) / length
    
    k = Young * Area / length
    rigimat = np.zeros((4,4), dtype=float)
    rigimat[0, 0] = k * cos_alpha ** 2
    rigimat[0, 1] = k * cos_alpha * sin_alpha
    rigimat[0, 2] = -k * cos_alpha ** 2
    rigimat[0, 3] = -k * cos_alpha * sin_alpha
    rigimat[1, 0] = k*cos_alpha*sin_alpha
    rigimat[1, 1] = k * sin_alpha ** 2
    rigimat[1, 2] = -k * cos_alpha * sin_alpha
    rigimat[1, 3] = -k * sin_alpha ** 2
    rigimat[2, 0] = -k * cos_alpha ** 2
    rigimat[2, 1] = -k * cos_alpha * sin_alpha
    rigimat[2, 2] = k*cos_alpha** 2
    rigimat[2, 3] = k * cos_alpha * sin_alpha
    rigimat[3, 0] = -k * cos_alpha * sin_alpha
    rigimat[3, 1] = -k * sin_alpha ** 2
    rigimat[3, 2] = k * cos_alpha * sin_alpha
    rigimat[3, 3] = k * sin_alpha ** 2
    return rigimat

def stiffness_TR03(elem):
    mat_id = elem["MaterialId"] - 1
    material = globalvars.data["Materials"][mat_id]
    Dmat = dmat_Solids2D(material)

    Bmat, Area = Bmat_TR03(elem)
    
    DmatxB = np.matmul(Dmat, Bmat)
    thickness = 1.0
    if(material["Plane_Type"] == "Plane_Stress"):
        thickness =material["Thickness"]
    rigimat = np.matmul(Bmat.transpose(), DmatxB) * Area * thickness
    return rigimat


def stiffness_QU04(elem):
    ProblemType = globalvars.data["ProblemType"]
    mat_id = elem["MaterialId"] - 1
    material = globalvars.data["Materials"][mat_id]
    Dmat = dmat_Solids2D(material)
    thickness = 1.0
    if(material["Plane_Type"] == "Plane_Stress"):
        thickness = material["Thickness"]
    Dmat *= thickness
    return stiffnessMatCalc(elem, "QU04", ProblemType, Dmat)

def stiffnessMatCalc(elem, ElemType, ProblemType, Dmat):
    ngauss = Set_Ngauss(ElemType)
    Nodes = globalvars.data["Mesh"]["Nodes"]
    ndof = GiveNdof(ElemType, ProblemType)
    nnodes = GiveNnodes(ElemType)
    size = nnodes * ndof
    rigimat = np.zeros((size,size), dtype=float)
    pos_pg, W = GaussQuadrature(ElemType)

    for igauss in range(ngauss):
        det_Jac, derivCart = derivCartesian(elem, ElemType, Nodes, pos_pg[igauss])
        Bmat = Calc_Bmat(ElemType, derivCart, ProblemType)
        DxBmat = np.matmul(Dmat, Bmat)
        rigimat_pg = np.matmul(Bmat.transpose(), DxBmat) * det_Jac * W[igauss]
        rigimat += rigimat_pg
    return rigimat


def assamk(doflist, rigimat):
  ndof = len(doflist)
  for idofn in range(ndof):
      ipos = doflist[idofn]
      for jdofn in range(ndof):
        jpos = doflist[jdofn]
        globalvars.astiff[ipos, jpos] += rigimat[idofn, jdofn]


def assamf(doflist, fvect):
  ndof = len(doflist)
  for idofn in range(ndof):
      ipos = doflist[idofn]
      globalvars.asload[ipos] += fvect[idofn]



import globalvars
import numpy as np

def GiveNdof(ElemType, ProblemType):
  if(ProblemType == "Structural_Mechanics"):
    if(ElemType == "BAR02" or ElemType == "BAR03"):
      ndof=1
    elif(ElemType == "TRUSS02" or ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
      ndof=2
  elif (ProblemType == "Thermal"):
    if(ElemType == "BAR02" or ElemType == "BAR03"):
      ndof=1
    elif(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
      ndof = 2
  elif (ProblemType == "Electrical"):
    if(ElemType == "BAR02" or ElemType == "BAR03"):
      ndof=1
  return ndof

def GiveNComp(ElemType, ProblemType):
  if(ProblemType == "Structural_Mechanics"):
    if(ElemType == "BAR02" or ElemType == "BAR03" or ElemType == "TRUSS02"):
        ncomp=1
    elif(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
        ncomp = 3
  elif (ProblemType == "Thermal"):
    if(ElemType == "BAR02" or ElemType == "BAR03" or ElemType == "TRUSS02"):
      ncomp=1
    elif(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
      ncomp = 2
  elif (ProblemType == "Electrical"):
    if(ElemType == "BAR02" or ElemType == "BAR03"):
      ncomp=1
  return ncomp

def initialization():
    mesh_type = globalvars.data["Mesh"]["ElemType"]
    ProblemType = globalvars.data["ProblemType"]
    globalvars.ndof = GiveNdof(mesh_type, ProblemType)
    globalvars.ncomp = GiveNComp(mesh_type, ProblemType)
    num_nodes = len(globalvars.data["Mesh"]["Nodes"])

    globalvars.madgln = np.full((num_nodes, globalvars.ndof), -2, dtype=int)

    nglpr = 0
    for ipres in range(len(globalvars.data["Constraints"])):
        id_node = globalvars.data["Constraints"][ipres]["Node"] - 1
        for igl in range(globalvars.ndof):
            if(globalvars.data["Constraints"][ipres]["Activation"][igl]):
                nglpr=nglpr+1
                globalvars.madgln[id_node, igl] = -1  #Flag value
    
    #Calculate madgln
    globalvars.neq = 0
    for elem in globalvars.data["Mesh"]["Elements"]:
      conec_list = elem["Connectivities"]
      for inode in range(len(conec_list)):
        id_node = conec_list[inode]-1
        for igl in range(globalvars.ndof):
          if (globalvars.madgln[id_node,igl] == -2):
            globalvars.madgln[id_node, igl] = globalvars.neq
            globalvars.neq += 1
            
    for ipres in range(len(globalvars.data["Constraints"])):
        id_node = globalvars.data["Constraints"][ipres]["Node"] - 1
        for igl in range(globalvars.ndof):
            if(globalvars.data["Constraints"][ipres]["Activation"][igl]):
                globalvars.madgln[id_node, igl] = globalvars.neq
                globalvars.neq += 1

    #Number of unknows
    globalvars.num_unknows = globalvars.neq - nglpr
    globalvars.astiff = np.zeros((globalvars.neq, globalvars.neq), dtype=float)
    globalvars.asload = np.zeros((globalvars.neq), dtype=float)
    globalvars.u_vec = np.zeros((globalvars.neq), dtype=float)

    num_known = globalvars.neq - globalvars.num_unknows
    globalvars.u_known = np.zeros((num_known), dtype=float)
    for ipres in range(len(globalvars.data["Constraints"])):
        id_node = globalvars.data["Constraints"][ipres]["Node"] - 1
        for igl in range(globalvars.ndof):
            if (globalvars.data["Constraints"][ipres]["Activation"][igl]):
                idire = globalvars.madgln[id_node, igl] - globalvars.num_unknows
                globalvars.u_known[idire] = globalvars.data["Constraints"][ipres]["Values"][igl]
    

    




import globalvars
import numpy as np
import json
from initialization import initialization
from solids_utils import Smooth_Strains, Smooth_Stresses, Smooth_Forces, GiveNComp
import matplotlib.pyplot as plt
import matplotlib.tri as tri
from plot_solids import plot_contour
from plot_trusses import plotMesh, plotDeformed, plotElemIntFluxes, plotNodalBarResult

#Set plotting defaults
gray = '#757575'
plt.rcParams['image.cmap'] = "rainbow"
plt.rcParams["mathtext.fontset"] = "cm"
plt.rcParams["text.color"] = gray
plt.rcParams["font.size"] = 10
plt.rcParams["xtick.color"] = gray
plt.rcParams["ytick.color"] = gray
plt.rcParams["axes.labelcolor"] = gray
plt.rcParams["axes.edgecolor"] = gray
plt.rcParams["axes.spines.right"] = False
plt.rcParams["axes.spines.top"] = False

def postprocess():
    ElemType = globalvars.data["Mesh"]["ElemType"]
    if("Show_displacements" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_displacements"]):
        plotDisplacements()

    if(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
        if("Show_strains" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_strains"]):
            plotStrains()
        
        if("Show_stresses" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_stresses"]):
            plotStresses()

        if("Show_forces" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_forces"]):
            plotForces()
    elif (ElemType == "BAR02" or ElemType == "BAR03" or ElemType == "TRUSS02"):
        plt.rcParams["axes.spines.right"] = True
        plt.rcParams["axes.spines.top"] = True
        plotMesh()
        if("Show_forces" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_forces"]):
            plotElemIntFluxes("Axial Forces", "N")

        if ("Show_temperatures" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_temperatures"]):
            plotTemperatures()

        if ("Show_voltage" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_voltage"]):
            plotVoltage()

        if("Show_thermal_fluxes" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_thermal_fluxes"]):
            plotElemIntFluxes("Thermal Fluxes", "q (W)")

        if("Show_current_intensity" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_current_intensity"]):
            plotElemIntFluxes("Current Intensity", "I (A)")                


    if("Show_reactions" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_reactions"]):
        writeReactions()

    resultsFile = globalvars.dataFileName.replace(".json", ".res.json")
    with open(resultsFile, 'w') as json_file:
        json.dump(globalvars.results, json_file, indent = 4)
    
    plt.show()

def plotDisplacements():
    mesh = globalvars.data["Mesh"]
    ElemType = mesh["ElemType"]
    ndof = globalvars.ndof
    nodal_disp_x = []
    nodal_disp_y = []
    globalvars.results["Displacements"] = []
    for inode in range(len(globalvars.data["Mesh"]["Nodes"])):
        idire_x = globalvars.madgln[inode, 0]
        disp_x = globalvars.u_vec[idire_x]
        nodal_disp_x.append(disp_x)

        if(ndof == 1):
            nodal_disp_res = {
                "Node": inode+1,
                "Disp_x": "{:10.4e}".format(disp_x),
            }
        elif(ndof == 2):
            idire_y = globalvars.madgln[inode, 1]
            disp_y = globalvars.u_vec[idire_y]
            nodal_disp_y.append(disp_y)
            nodal_disp_res = {
                "Node": inode+1,
                "Disp_x": "{:10.4e}".format(disp_x),
                "Disp_y": "{:10.4e}".format(disp_y)
            }

        globalvars.results["Displacements"].append(nodal_disp_res)
    
    if(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
        plot_contour(mesh, nodal_disp_x, "Displacements x", r"$u_x$ (m)")
        plot_contour(mesh, nodal_disp_y, "Displacements y", r"$u_y$ (m)")
    elif (ElemType == "BAR02" or ElemType == "BAR03"):
        plt.rcParams["axes.spines.right"] = True
        plt.rcParams["axes.spines.top"] = True
        plotNodalBarResult("Displacements", r"$u_x$ (m)", nodal_disp_x)
    elif (ElemType == "TRUSS02"):
        plt.rcParams["axes.spines.right"] = True
        plt.rcParams["axes.spines.top"] = True
        plotNodalBarResult("Horizontal displacements", r"$u_x$ (m)", nodal_disp_x)
        plotNodalBarResult("Vertical displacements", r"$u_y$ (m)", nodal_disp_y)
        if ("Show_deformed" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Show_deformed"]):
            plotDeformed()


def plotStrains():
    mesh = globalvars.data["Mesh"]
    ncomp = globalvars.ncomp
    nodal_Epsilon = Smooth_Strains()

    globalvars.results["Strains"] = []
    for inode in range(len(globalvars.data["Mesh"]["Nodes"])):
        if (ncomp == 1):
            nodal_strains_res = {
                "Node": inode+1,
                "Epsilon_x": "{:10.4e}".format(nodal_Epsilon[inode,0]),
            }
        elif (ncomp == 3):
            nodal_strains_res = {
                "Node": inode+1,
                "Epsilon_x": "{:10.4e}".format(nodal_Epsilon[inode,0]),
                "Epsilon_y": "{:10.4e}".format(nodal_Epsilon[inode,1]),
                "Gama_xy":   "{:10.4e}".format(nodal_Epsilon[inode,2])
            }
        globalvars.results["Strains"].append(nodal_strains_res) 

    if(ncomp == 3):
        plot_contour(mesh, nodal_Epsilon[:,0], "Strains Epsilon_x", r"$\epsilon_{x}$")
        plot_contour(mesh, nodal_Epsilon[:,1], "Strains Epsilon_y", r"$\epsilon_{y}$")
        plot_contour(mesh, nodal_Epsilon[:,2], "Strains Gama_xy", r"$\gamma_{xy}$")


def plotStresses():
    mesh = globalvars.data["Mesh"]
    ncomp = globalvars.ncomp
    nodal_Sigma = Smooth_Stresses()

    globalvars.results["Stresses"] = []
    for inode in range(len(globalvars.data["Mesh"]["Nodes"])):
        if (ncomp == 1):
            nodal_stresses_res = {
                "Node": inode+1,
                "Sigma_x": "{:10.4e}".format(nodal_Sigma[inode,0]),
            }
        elif (ncomp == 3):
            nodal_stresses_res = {
                "Node": inode+1,
                "Sigma_x": "{:10.4e}".format(nodal_Sigma[inode,0]),
                "Sigma_y": "{:10.4e}".format(nodal_Sigma[inode,1]),
                "Tau_xy":  "{:10.4e}".format(nodal_Sigma[inode,2])
            }
        globalvars.results["Stresses"].append(nodal_stresses_res)        

    if(ncomp == 3):
        plot_contour(mesh, nodal_Sigma[:,0], "Stresses Sigma_x", r"$\sigma_{x}$")
        plot_contour(mesh, nodal_Sigma[:,1], "Stresses Sigma_y", r"$\sigma_{y}$")
        plot_contour(mesh, nodal_Sigma[:, 2], "Stresses Tau_xy", r"$\tau_{xy}$")
        

def plotForces():
    mesh = globalvars.data["Mesh"]
    ncomp = globalvars.ncomp
    nodal_Forces = Smooth_Forces()

    globalvars.results["Forces"] = []
    for inode in range(len(globalvars.data["Mesh"]["Nodes"])):
        if (ncomp == 1):
            nodal_forces_res = {
                "Node": inode+1,
                "N_x": "{:10.4e}".format(nodal_Forces[inode,0]),
            }
        elif (ncomp == 3):
            nodal_forces_res = {
                "Node": inode+1,
                "N_x": "{:10.4e}".format(nodal_Forces[inode,0]),
                "N_y": "{:10.4e}".format(nodal_Forces[inode,1]),
                "N_xy":  "{:10.4e}".format(nodal_Forces[inode,2])
            }
        globalvars.results["Forces"].append(nodal_forces_res)        

    if(ncomp == 3):
        plot_contour(mesh, nodal_Forces[:,0], "Internal Forces N_x", r"$N_{x}$")
        plot_contour(mesh, nodal_Forces[:,1], "Internal Forces N_y", r"$N_{y}$")
        plot_contour(mesh, nodal_Forces[:, 2], "Internal Forces N_xy", r"$N_{xy}$")

def plotTemperatures():
    mesh = globalvars.data["Mesh"]
    ElemType = mesh["ElemType"]
    nodal_temp = []
    globalvars.results["Temperatures"] = []
    for inode in range(len(globalvars.data["Mesh"]["Nodes"])):
        idire = globalvars.madgln[inode, 0]
        temp = globalvars.u_vec[idire]
        nodal_temp.append(temp)

        nodal_temp_res = {
            "Node": inode+1,
            "Temp": "{:10.4e}".format(temp),
        }

        globalvars.results["Temperatures"].append(nodal_temp_res)
    
    if(ElemType == "TR03" or ElemType == "TR06" or ElemType == "QU04" or ElemType == "QU08" or ElemType == "QU09"):
        plot_contour(mesh, nodal_temp, "Temperatures", r"$Temp$ ($^\circ$C)")
    elif (ElemType == "BAR02" or ElemType == "BAR03"):
        plotNodalBarResult("Temperatures", r"$Temp$ ($^\circ$C)", nodal_temp)
        
def plotVoltage():
    mesh = globalvars.data["Mesh"]
    ElemType = mesh["ElemType"]
    nodal_temp = []
    globalvars.results["Voltage"] = []
    for inode in range(len(globalvars.data["Mesh"]["Nodes"])):
        idire = globalvars.madgln[inode, 0]
        temp = globalvars.u_vec[idire]
        nodal_temp.append(temp)

        nodal_temp_res = {
            "Node": inode+1,
            "Voltage": "{:10.4e}".format(temp),
        }

        globalvars.results["Voltage"].append(nodal_temp_res)
    
    if (ElemType == "BAR02" or ElemType == "BAR03"):
        plotNodalBarResult("Voltage", r"$Voltage$ (V)", nodal_temp)

def writeReactions():
    globalvars.results["Reactions"] = []
    for ipres in range(len(globalvars.data["Constraints"])):
        node = globalvars.data["Constraints"][ipres]["Node"]
        id_node = node - 1
        if (globalvars.data["ProblemType"] == "Structural_Mechanics"):
            react_results_node = {
                "Node": node,
                "Rx": "{:10.4e}".format(0.0),
                "Ry": "{:10.4e}".format(0.0)
            }
            for igl in range(globalvars.ndof):
                if(globalvars.data["Constraints"][ipres]["Activation"][igl]):
                    idire = globalvars.madgln[id_node, igl] - globalvars.num_unknows
                    if(igl == 0):
                        react_results_node["Rx"] = "{:10.4e}".format(globalvars.react_vec[idire])
                    elif(igl == 1):
                        react_results_node["Ry"] ="{:10.4e}".format(globalvars.react_vec[idire])
        elif (globalvars.data["ProblemType"] == "Thermal"):
            idire = globalvars.madgln[id_node, 0] - globalvars.num_unknows
            react_results_node = {
                "Node": node,
                "Flux": "{:10.4e}".format(globalvars.react_vec[idire])
            }
        elif (globalvars.data["ProblemType"] == "Electrical"):
            idire = globalvars.madgln[id_node, 0] - globalvars.num_unknows
            react_results_node = {
                "Node": node,
                "Current_Intensity": "{:10.4e}".format(globalvars.react_vec[idire])
            }
        
        globalvars.results["Reactions"].append(react_results_node)
        




import numpy as np
import math
import matplotlib.pyplot as plt
import matplotlib as mpl
import globalvars
from initialization import initialization
from bars_utils import ElemTrusses_IntForces, ElemBars_IntFluxes
from plot_solids import plot_contour

def plotMesh(figTitle="Mesh"):
    window = plt.figure(figTitle)
    nodes = globalvars.data["Mesh"]["Nodes"]
    coords = np.zeros((len(nodes), 2), dtype=float)
    for i, node in enumerate(nodes):
        coords[i, 0] = node["x"]
        coords[i, 1] = node["y"]

    elements = globalvars.data["Mesh"]["Elements"]
    elemList= np.zeros((len(elements), 2), dtype=int)
    for i, elem in enumerate(elements):
        elemList[i, :] = elem["Connectivities"]

    fig, ax = plot_nodes(coords, color='k')
    fig, ax = plot_elems(coords, elemList, fig, ax, color='C7')
    label2d(coords, elemList, ax)
    window.tight_layout()

def plotDeformed():
    window = plt.figure("Deformed")
    nodes = globalvars.data["Mesh"]["Nodes"]
    coords = np.zeros((len(nodes), 2), dtype=float)
    scale_disp = globalvars.scale_disp
    if ("Deformed_scale" in globalvars.data["Postprocess"] and globalvars.data["Postprocess"]["Deformed_scale"]):
        scale_disp = globalvars.data["Postprocess"]["Deformed_scale"]
    for i, node in enumerate(nodes):
        idire_x = globalvars.madgln[i, 0]
        idire_y = globalvars.madgln[i, 1]
        disp_x = globalvars.u_vec[idire_x]
        disp_y = globalvars.u_vec[idire_y]
        coords[i, 0] = node["x"] + scale_disp*disp_x
        coords[i, 1] = node["y"] + scale_disp*disp_y

    elements = globalvars.data["Mesh"]["Elements"]
    E= np.zeros((len(elements), 2), dtype=int)
    for i, elem in enumerate(elements):
        E[i, :] = elem["Connectivities"]

    fig, ax = plot_nodes(coords, color='k')
    fig, ax = plot_elems(coords, E, fig, ax, color='C7')
    label2d(coords, E, ax)
    window.tight_layout()
    
def plotElemIntFluxes(figTitle, plotTitle):
    if(figTitle =="Axial Forces"):
        ElemIntFluxes = ElemTrusses_IntForces()
    elif (figTitle == "Thermal Fluxes" or figTitle == "Current Intensity"):
        ElemIntFluxes = ElemBars_IntFluxes()
    
    window = plt.figure(figTitle)
    nodes = globalvars.data["Mesh"]["Nodes"]
    coords = np.zeros((len(nodes), 2), dtype=float)
    for i, node in enumerate(nodes):
        coords[i, 0] = node["x"]
        coords[i, 1] = node["y"]

    elements = globalvars.data["Mesh"]["Elements"]
    elemList= np.zeros((len(elements), 2), dtype=int)
    for i, elem in enumerate(elements):
        elemList[i, :] = elem["Connectivities"]

    fig, ax = plot_nodes(coords, color='k')

    plt.title(plotTitle)
    contour_limit = np.zeros((2), dtype=float)
    min_flux = ElemIntFluxes.min()
    max_flux = ElemIntFluxes.max()
    if (abs(min_flux - max_flux) < 1.0e-10):
        if (min_flux < 0):
            max_flux = 0.0
        if (max_flux > 0):
            min_flux = 0.0
    contour_limit[0] = min_flux
    contour_limit[1] = max_flux

    con = plotTitle, ElemIntFluxes, contour_limit
    fig, ax = plot_elems(coords, elemList, fig, ax, color='C7', contour=con)
    labelBarValue(coords, elemList, ElemIntFluxes, ax)
    window.tight_layout()

def plot_nodes(coords, color='k', size=10):
    plt.scatter(coords[:,0],coords[:,1],marker='o',s=15*size,color=color,zorder=30)
    fig = plt.gcf()
    ax = plt.gca()
    return fig, ax

def plot_elems(coords, E, fig=None, ax=None, color='C0', contour=None, lim_scale=1.2):
 
    if contour is not None:
        clabel = contour[0]
        contour_lim = contour[2]
        contour = contour[1]
        
    for k,e in enumerate(E):
        NA = int(e[-2])
        NE = int(e[-1])
        
        if contour is not None:
            # build up colormap and normalizer
            colormap = plt.get_cmap('bwr', 10)
            colormap.set_over('0.95')
            colormap.set_under('0.6')
            
            if contour_lim is None:
                norm = mpl.colors.Normalize(vmin=min(contour), vmax=max(contour))
            else:
                norm = mpl.colors.Normalize(vmin=contour_lim[0], vmax=contour_lim[1])
            color = colormap(norm(contour[k]))
            
            # create a ScalarMappable and initialize a data structure
            s_m = mpl.cm.ScalarMappable(cmap=colormap, norm=norm)
            s_m.set_array([])
            
        if contour is not None:
                _, = plt.plot([coords[NA-1][0],coords[NE-1][0]],
                                    [coords[NA-1][1],coords[NE-1][1]],color='C7',zorder=1,linewidth=9.0)
        _, = plt.plot([coords[NA-1][0],coords[NE-1][0]],
                            [coords[NA-1][1],coords[NE-1][1]],color=color,zorder=2,linewidth=5.0)
        
    if contour is not None:
        plt.colorbar(s_m, ticks=np.linspace(contour_lim[0],contour_lim[1],11))
        plt.ylabel('CONTOUR = '+clabel)
        
    m = lim_scale*max(abs(plt.xlim()[0]),
                    abs(plt.xlim()[1]),
                    abs(plt.ylim()[0]),
                    abs(plt.ylim()[1]))
        
    if lim_scale < 0:
        m = -lim_scale
    else:
        alimx = (plt.xlim()[1] - plt.xlim()[0])/2
        alimy = (plt.ylim()[1] - plt.ylim()[0])/2
        mlimx = (plt.xlim()[1] + plt.xlim()[0])/2
        mlimy = (plt.ylim()[1] + plt.ylim()[0])/2
        alim = lim_scale*max(alimx,alimy)
        minx, maxx = (mlimx-alim, mlimx+alim)
        miny, maxy = (mlimy-alim, mlimy+alim)
        if abs(lim_scale) < 100:
            plt.xlim(minx, maxx)
            plt.ylim(miny, maxy)

    plt.gca().set_aspect('equal')
    if lim_scale < 0 and abs(lim_scale) < 100:
        plt.xlim(-m,m)
        plt.ylim(-m,m)

    plt.gcf().set_size_inches(8,6)
    
    return fig, ax

def label2d(coords, elemList, axes):
    """Plot the nodes and element label
    """
    x_min = coords[:,0].min()
    x_max = coords[:,0].max()
    y_min = coords[:,1].min()
    y_max = coords[:, 1].max()
    max_size = max((x_max - x_min), (y_max - y_min))
    offset_fact = 0.02
    offset_x = offset_fact * max_size
    offset_y = offset_fact * max_size
    if (x_min == x_max):
        offset_y = 0.0
    if (y_min == y_max):
        offset_x = 0.0

    for inode, node in enumerate(coords):
        x_pos = node[0] + offset_x
        y_pos = node[1] + offset_y
        axes.text( x_pos , y_pos, str(inode+1), color='b', size=10)

    for ele, con in enumerate(elemList):
        xm = (coords[con[0]-1][0] + coords[con[1]-1][0]) / 2 + offset_x
        ym = (coords[con[0]-1][1] + coords[con[1]-1][1]) / 2 + offset_y
        axes.text(xm, ym, str(ele + 1), color='g', size=10)
        
def labelBarValue(coords, elemList, valueList, axes):
    """Plot text with bar value
    """
    x_min = coords[:,0].min()
    x_max = coords[:,0].max()
    y_min = coords[:,1].min()
    y_max = coords[:, 1].max()
    max_size = max((x_max - x_min), (y_max - y_min))
    offset_fact = 0.02
    offset_x = offset_fact * max_size
    offset_y = offset_fact * max_size
    if (x_min == x_max):
        offset_y = 0.0
    if (y_min == y_max):
        offset_x = 0.0

    for ielem, con in enumerate(elemList):
        xm = (coords[con[0]-1][0] + coords[con[1]-1][0]) / 2 + offset_x
        ym = (coords[con[0] - 1][1] + coords[con[1] - 1][1]) / 2 + offset_y
        #valueFormatted = round(valueList[ielem],4)
        valueFormatted =  '{0:.5g}'.format(valueList[ielem])
        axes.text(xm, ym, valueFormatted, color='g', size=12)

def plotNodalBarResult(figTitle, plotTitle, nodalValues):
    window = plt.figure(figTitle)
    quadsMesh = createFictitiousQuadsMeshFromBars(globalvars.data["Mesh"])
    quadNodalValues = createQuadMeshNodalValues(nodalValues, globalvars.data["Mesh"]["Elements"])
    nodes = globalvars.data["Mesh"]["Nodes"]
    coords = np.zeros((len(nodes), 2), dtype=float)
    for i, node in enumerate(nodes):
        coords[i, 0] = node["x"]
        coords[i, 1] = node["y"]
    plot_nodes(coords, color='k')
    plot_contour(quadsMesh, quadNodalValues, figTitle, plotTitle)
    window.tight_layout()

def createFictitiousQuadsMeshFromBars(barMesh):
    barNodesList = barMesh["Nodes"]
    barElemsList = barMesh["Elements"]

    x_vec = np.zeros((len(barNodesList)), dtype=float)
    y_vec = np.zeros((len(barNodesList)), dtype=float)
    for inode, node in enumerate(barNodesList):
        x_vec[inode] = node["x"]
        y_vec[inode] = node["y"]
    x_min = x_vec.min()
    x_max = x_vec.max()
    y_min = y_vec.min()
    y_max = y_vec.max()
    max_size = max((x_max - x_min), (y_max - y_min))
    elementHeight = 0.04 * max_size

    quadsMesh = {
        "ElemType": "QU04",
        "Nodes": [],
        "Elements": []
    }

    for ielem, barElem in enumerate(barElemsList):
        node1 = barElem["Connectivities"][0] - 1
        x1 = barNodesList[node1]["x"]
        y1 = barNodesList[node1]["y"]
        node2 = barElem["Connectivities"][1] - 1
        x2 = barNodesList[node2]["x"]
        y2 = barNodesList[node2]["y"]
        l = math.sqrt((x2 - x1)** 2 + (y2 - y1)** 2)
        cos_alpha = (x2 - x1) / l
        sin_alpha = (y2 - y1) / l
        nx = -sin_alpha
        ny = cos_alpha

        materialId = barElem["MaterialId"]
        quadElem = {
            "Connectivities": [],
            "MaterialId": materialId
        }

        node_quad_1 = {
            "x": x1 - nx*elementHeight / 2,
            "y": y1 - ny*elementHeight / 2
        }
        quadsMesh["Nodes"].append(node_quad_1)

        node_quad_2 = {
            "x": x2 - nx*elementHeight / 2,
            "y": y2 - ny*elementHeight / 2
        }
        quadsMesh["Nodes"].append(node_quad_2)

        
        node_quad_3 = {
            "x": x2 + nx*elementHeight / 2,
            "y": y2 + ny*elementHeight / 2
        }
        quadsMesh["Nodes"].append(node_quad_3)

        node_quad_4 = {
            "x": x1 + nx*elementHeight / 2,
            "y": y1 + ny*elementHeight / 2
        }
        quadsMesh["Nodes"].append(node_quad_4)

        inode_quad_1 = ielem * 4 + 1
        quadElem["Connectivities"].append(inode_quad_1)
        inode_quad_2 = ielem * 4 + 2
        quadElem["Connectivities"].append(inode_quad_2)
        inode_quad_3 = ielem * 4 + 3
        quadElem["Connectivities"].append(inode_quad_3)
        inode_quad_4 = ielem * 4 + 4
        quadElem["Connectivities"].append(inode_quad_4)

        quadsMesh["Elements"].append(quadElem)
    
    return quadsMesh

def createQuadMeshNodalValues(barNodalValues, barElemsList):
    quadNodalValues = np.zeros((4*len(barElemsList)),dtype=float)
    for ielem, barElem in enumerate(barElemsList):
        nodeBar_1 = barElem["Connectivities"][0] - 1
        valueBarNode1 = barNodalValues[nodeBar_1]
        nodeBar_2 = barElem["Connectivities"][1] - 1
        valueBarNode2 = barNodalValues[nodeBar_2]

        nodeQuad_1 = ielem * 4
        quadNodalValues[nodeQuad_1] = valueBarNode1

        nodeQuad_2 = ielem * 4 + 1
        quadNodalValues[nodeQuad_2] = valueBarNode2

        nodeQuad_3 = ielem * 4 + 2
        quadNodalValues[nodeQuad_3] =valueBarNode2

        nodeQuad_4 = ielem * 4 + 3
        quadNodalValues[nodeQuad_4] = valueBarNode1

    return quadNodalValues



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.tri as tri

# converts quad elements into triangle elements
def quads_to_triangles(quadsMesh):
    trianglesMesh = {
        "ElemType": "TR03",
        "Nodes": quadsMesh["Nodes"],
        "Elements": []
    }

    for quadElem in quadsMesh["Elements"]:
        node = np.empty((4),dtype=int)
        node[:4] =  quadElem["Connectivities"][:4]
        triangleElem1 = {
            "Connectivities": [node[0], node[1], node[2]],
            "MaterialId": quadElem["MaterialId"]
        }
        trianglesMesh["Elements"].append(triangleElem1)

        triangleElem2 = {
            "Connectivities": [node[2], node[3], node[0]],
            "MaterialId": quadElem["MaterialId"]
        }
        trianglesMesh["Elements"].append(triangleElem2)
    return trianglesMesh


# plots a finite element mesh
def plot_fem_mesh(mesh):
    for element in mesh["Elements"]:
        x = []
        y = []
        for node in element["Connectivities"]:
            id_node = node - 1
            x.append(mesh["Nodes"][id_node]["x"])
            y.append(mesh["Nodes"][id_node]["y"])
        plt.fill(x, y, edgecolor='black', fill=False)

def plot_contour(mesh, nodal_values, figTitle, plotTitle):
    mesh_triangulation = mesh
    if (mesh["ElemType"] == "QU04"):
        mesh_triangulation = quads_to_triangles(mesh)
    
    x_coords = []
    y_coords = []
    for node in mesh_triangulation["Nodes"]:
        x_coords.append(node["x"])
        y_coords.append(node["y"])

    connectivities_list = []
    for element in mesh_triangulation["Elements"]:
        connectivities_elem = []
        for node in element["Connectivities"]:
            id_node = node - 1
            connectivities_elem.append(id_node)
        connectivities_list.append(connectivities_elem)

    # create an unstructured triangular grid instance
    triangulation = tri.Triangulation(x_coords, y_coords, connectivities_list)

    # plot the contours
    plt.figure(figTitle)
    plot_fem_mesh(mesh)
    plt.tricontourf(triangulation, nodal_values, 12)

    # show
    plt.title(plotTitle)
    plt.colorbar(orientation='vertical')
    plt.axis('equal')



import numpy as np
import math
import globalvars
from initialization import initialization
from dmat import giveLocalStiffness

def ElemTrusses_IntForces():
    globalvars.results["ElemIntForces"] = []
    mesh = globalvars.data["Mesh"]
    nodeList = mesh["Nodes"]
    elemList = mesh["Elements"]
    ElemIntForces = np.zeros(len(elemList), dtype=float)
    node = np.array([0, 0])
    idire = np.zeros((2), dtype=int)
    global_disp = np.zeros((2, 2), dtype=float)
    local_disp = np.zeros((2), dtype=float)
    for ielem, elem in enumerate(elemList):
        mat_id = elem["MaterialId"] - 1
        material = globalvars.data["Materials"][mat_id]
        node1 = elem["Connectivities"][0] - 1
        x1 = nodeList[node1]["x"]
        y1 = nodeList[node1]["y"]
        node2 = elem["Connectivities"][1] - 1
        x2 = nodeList[node2]["x"]
        y2 = nodeList[node2]["y"]
        length = math.sqrt((x2 - x1)** 2 + (y2 - y1)** 2)
        cos_alpha = (x2 - x1) / length
        sin_alpha = (y2 - y1) / length
        k = giveLocalStiffness(material, length)

        for inode in range(2):
            node[inode] = elem["Connectivities"][inode] - 1
            for igl in range(globalvars.ndof):
                idire = globalvars.madgln[node[inode], igl]
                global_disp[inode, igl] = globalvars.u_vec[idire]
            local_disp[inode]=global_disp[inode,0]*cos_alpha + global_disp[inode,1]*sin_alpha
        
        axialForce = k * (local_disp[1] - local_disp[0])
        ElemIntForces[ielem] = axialForce
        axialForcesFormatted = "{:10.4e}".format(axialForce)
        globalvars.results["ElemIntForces"].append(axialForcesFormatted)

    return ElemIntForces

def ElemBars_IntFluxes():
    globalvars.results["ElemIntFluxes"] = []
    mesh = globalvars.data["Mesh"]
    nodeList = mesh["Nodes"]
    elemList = mesh["Elements"]
    ElemIntFluxes = np.zeros(len(elemList), dtype=float)
    node = np.array([0, 0])
    idire = np.zeros((2), dtype=int)
    var_vec = np.zeros((2), dtype=float)
    for ielem, elem in enumerate(elemList):
        mat_id = elem["MaterialId"] - 1
        material = globalvars.data["Materials"][mat_id]
        node1 = elem["Connectivities"][0] - 1
        x1 = nodeList[node1]["x"]
        y1 = nodeList[node1]["y"]
        node2 = elem["Connectivities"][1] - 1
        x2 = nodeList[node2]["x"]
        y2 = nodeList[node2]["y"]
        length = math.sqrt((x2 - x1)** 2 + (y2 - y1)** 2)
        k = giveLocalStiffness(material, length)

        for inode in range(2):
            node[inode] = elem["Connectivities"][inode] - 1
            idire = globalvars.madgln[node[inode], 0]
            var_vec[inode] = globalvars.u_vec[idire]
        
        flux = k * (var_vec[1] - var_vec[0])
        ElemIntFluxes[ielem] = flux
        fluxFormatted = "{:10.4e}".format(flux)
        globalvars.results["ElemIntFluxes"].append(fluxFormatted)

    return ElemIntFluxes






import numpy as np

data = {}
neq = 0
ndof = 1
ncomp = 1
num_unknows = 0
madgln = []
astiff = []
asload = []
u_vec = []
u_known = []
react_vec = []
results = {}
Eps_elem_vec = [[], [], []]
Sigma_elem_vec = [[], [], []]
scale_disp = 1.0
dataFileName = ""



import json
import os
import globalvars

def read_data():
    
    print("Input the name of the json data file with its relative path (e.g. project/example.json).")
    fileName = input("Press enter when the file is called data.json and is located in the current folder: ")
    if(fileName == ""):
        base_dir = os.getcwd()
        fileName = os.path.join(base_dir, "data.json")
    globalvars.dataFileName = fileName
    with open(fileName, "r") as read_file:
        globalvars.data = json.load(read_file)




from distutils.core import setup
setup(
  name = 'sopy_fem',
  packages = ['sopy_fem'], # this must be the same as the name above
  version = '0.0.1',
  description = 'An academic program for finite elments analysis of 2D solids and structures',
  author = 'Daniel Di Capua',
  author_email = 'dicapua67@gmail.com',
  url = 'https://github.com/DanielDiCapua/sopy_fem.git', # use the URL to the github repo
  download_url = 'https://github.com/DanielDiCapua/sopy_fem.git/tarball/0.0.1',
  keywords = ['developing', 'example'],
  classifiers = [],
)


import globalvars
import numpy as np
from initialization import initialization

def solve():
    n = globalvars.num_unknows
    neq = globalvars.neq

    #Incluiding the effect of the known displacements u_known in load vector bu
    bu = globalvars.asload[:n]
    Auk = globalvars.astiff[:n, n:]
    bu -= np.matmul(Auk, globalvars.u_known)
    num_known = neq - n

    #Solve system of equations => Auu·u_unknows = bu
    Auu = globalvars.astiff[:n,:n]
    u_unknows = np.linalg.solve(Auu, bu)

    globalvars.u_vec[:n] = u_unknows[:n]
    globalvars.u_vec[n:neq] = globalvars.u_known[:num_known]

    #Reactions calculation
    bk = globalvars.asload[n:]
    Aku = globalvars.astiff[n:, :n]
    Akk = globalvars.astiff[n:, n:]
    globalvars.react_vec = np.matmul(Aku, u_unknows) + np.matmul(Akk, globalvars.u_known) - bk



import globalvars
import numpy as np

def dmat_Solids2D(material):
    dmat = np.zeros((3, 3), dtype=float)
    Young = material["Young"]
    Poisson = material["Poisson"]
    G = Young/(2.0*(1.0+Poisson))

    if (material["Plane_Type"] == "Plane_Stress"):
        fact1 = Young/(1.0-Poisson**2)
        dmat[0, 0] = fact1
        dmat[0, 1] = fact1 * Poisson
        dmat[1, 0] = fact1 * Poisson
        dmat[1, 1] = fact1
        dmat[2,2] = G
    elif (material["Plane_Type"] == "Plane_Strain"):
        fact1 = Young * (1.0 - Poisson) / ((1.0 + Poisson) * (1 - 2.0 * Poisson))
        fact2= Young *  Poisson / ((1.0 + Poisson) * (1 - 2.0 * Poisson))
        dmat[0, 0] = fact1
        dmat[0, 1] = fact2
        dmat[1, 0] = fact2
        dmat[1, 1] = fact1
        dmat[2, 2] = G
        
    return dmat

def giveLocalStiffness(material, l):
    ProblemType = globalvars.data["ProblemType"]
    if (ProblemType == "Structural_Mechanics"):
        if("Young" in material and "Area" in material):
            Young = material["Young"]
            Area = material["Area"]
        else:
            raise Exception("The Area and Young definitions are mandatory for mechanics bars")
        k = Young * Area / l
    elif (ProblemType == "Thermal"):
        if("Thermal_Conductivity" in material and "Area" in material):
            Thermal_Cond = material["Thermal_Conductivity"]
            Area = material["Area"]
        else:
            raise Exception("The Area and Thermal_Conductivity definitions are mandatory for thermal bars")
        k = Thermal_Cond * Area / l
    elif (ProblemType == "Electrical"):
        if("Electrical_Conductivity" in material and "Area" in material):
            Electrical_Cond = material["Electrical_Conductivity"]
            Area = material["Area"]
        else:
            raise Exception("The Area and Electrical_Conductivity definitions are mandatory for electrical bars")
        k = Electrical_Cond * Area / l
    
    return k
        
        

        



    